! StartingPoints-1D Copyright (C) 2013 - 2016  Marco Spadini
! This file is part of StartingPoints-1D.

! StartingPoints-1D is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! StartingPoints-1D is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with StartingPoints-1D.  If not, see <http://www.gnu.org/licenses/>.
! 
! 
! About the author:
! Marco Spadini                         
! Dipartimento di Matematica e Informatica,
! Via S. Marta 3, 50139 Firenze,       
! ITALY 
! 
! web: http:\\www.dma.unifi.it\~spadini
! e-mail: marco.spadini@unifi.it
!


 module rk_scalar  !! Runge-Kutta methods, 1D problems with one parameter
   implicit none
   integer,parameter,public :: db = selected_real_kind(p=16,r=200)
   real(kind=db),parameter,public    :: atoll = 1.E-6_db, rtoll = 1.E-3_db, hmax = 0.5_db, hmin = 0.00125_db,&
                                        hinit = 10._db*hmin, mins = 0.2_db, maxs = 10._db
   real(kind=db),parameter,private :: a21=1._db/5._db,&
                                      a31=3._db/40._db, a32= 9._db/40._db,&
                                      a41=44._db/45._db, a42= -56._db/15._db, a43= 32._db/9._db,&
                                      a51=19372._db/6561._db, a52=-25360/2817._db, a53=64448._db/6561._db, &
                                                              a54=-212._db/729._db,&
                                      a61=9017._db/3168._db, a62=-355._db/33._db, a63=46723._db/5247._db, &
                                                              a64=49._db/174._db, a65=-5103._db/18656._db,&
                                      a71=35._db/384._db, a73=500._db/1113._db, a74=125._db/192._db, &
                                                              a75=-2187._db/6784._db, a76= 11._db/84._db,&
                                      c2=1._db/5._db, c3=3._db/10._db, c4=4._db/5._db, c5= 8._db/9._db,&
                                      b1=5179._db/57600._db, b3=7571._db/16695._db, b4=393._db/640._db, &
                                                              b5=-92097._db/339200._db,&
                                      b6=187._db/2100._db, b7=1._db/40._db, &
                                      e1 = 71._db/57600._db, e3 = -71._db/16695._db, e4 = 71._db/1920._db, &
                                                              e5 = -17253._db/339200._db, e6 = 22._db/525._db
   contains
   
   

   pure function d_RKDP_adaptive(f,x0,periodo,lambda) result(yy) 
       implicit none
       interface
         pure function f(t,lambda,xi)
           integer,parameter :: db = selected_real_kind(p=16,r=200)
           real(kind=db),intent(in) :: t,lambda
           real(kind=db),intent(in) :: xi
           real(kind=db) :: f
           end function f
         end interface
       real(kind=db),intent(IN) :: x0, lambda, periodo
       real(kind=db) :: k0, k1, k2, k3, k4, k5, k6, k7, k8, yy, yy_new, R, Rold, s, h, errass, t0
       
       Rold = atoll
       t0 = 0._db
       h = hinit
       yy = x0
       k0 =  f(t0,lambda,yy)
       
       mcycle: do !while ( t0 .lt. periodo )
         attempts: do
         k1 = k0
         k2 = f(t0+c2*h,lambda,yy+h*a21*k1)
         k3 = f(t0+c3*h,lambda,yy+h*(a31*k1+a32*k2))
         k4 = f(t0+c4*h,lambda,yy+h*(a41*k1+a42*k2+a43*k3))
         k5 = f(t0+c5*h,lambda,yy+h*(a51*k1+a52*k2+a53*k3+a54*k4))
         k6 = f(t0+h,lambda,yy+h*(a61*k1+a62*k2+a63*k3+a64*k4+a65*k5))
         k7 = f(t0+h,lambda,yy+h*(a71*k1+a73*k3+a74*k4+a75*k5+a76*k6))
         yy_new = yy+h*(k1*b1+k3*b3+k4*b3+k5*b5+k6*b6+k7*b7) !! 6th order approx
         !yy5 = yy+h*(k1*a71+k3*a73+k4*a74+k5*a75+k6*a76)    !! 5th order (not used)
         errass = abs(h*(k1*e1+k3*e3+k4*e4+k5*e5+k6*e6-k7*b7))
         R = max(atoll,errass/(atoll+rtoll*max(abs(yy_new),abs(yy))))            
         s = max(mins,min(maxs,0.9_db*R**(-0.17_db)*(Rold**0.04_db)) ) !! PI  control: Alpha=0.2, Beta = 0.04
         
         if ((R .lt. 1.) .or. (h .le. hmin))  then  !! accept step
           t0= t0 + h
           k0 = k7
           yy=yy_new
           h=min(s*h,hmax,periodo-t0)
           Rold = R
           exit attempts
         else                                       !! reject step
           h=max(min(s,1._db)*h,hmin)
           k0 = f(t0,lambda,yy)
           endif
           end do attempts
           
         if ( t0 .ge. periodo)   exit mcycle
         end do mcycle
       
       end function d_RKDP_adaptive



   pure function d_RKV56_adaptive(rhs,x0,periodo,lambda) result(yy) 
       implicit none
       interface
         pure function rhs(t,lambda,xi)
           integer,parameter :: db = selected_real_kind(p=16,r=200)
           real(kind=db),intent(in) :: t,lambda
           real(kind=db),intent(in) :: xi
           real(kind=db) :: rhs
           end function rhs
         end interface
       real(kind=db) :: yy
       real(kind=db),intent(IN) :: x0
       real(kind=db),intent(IN) :: lambda, periodo
       real(kind=db),parameter :: c1=4._db/15._db, c2=4._db/75._db, c3=16._db/75._db, &
                             d1=2._db/3._db, d2=5._db/6._db, d3=8._db/3._db, d4= 5._db/2._db, &
                             e1=165._db/64._db, e2=55._db/6._db, e3= 425._db/64, e4=85._db/96._db, &
                             f1=12._db/5._db, f2=4015._db/612._db, f3=11._db/36._db, f4=88._db/255._db, &
                             g1=8263._db/15000._db, g2=124._db/75._db, g3=643._db/680._db, &
                             g4=81._db/250._db, g5=2484._db/10625._db, &
                             h1=3501._db/1720._db, h2=300._db/43._db, h3=297275._db/52632._db, &
                             h4=319._db/2322._db, h5=24068._db/84065._db, h6=3850._db/26703._db,&
                             w1=125._db/17952._db, w2=12._db/1955._db, w3=3._db/44._db, &
                             w4=125._db/11592._db, w5=43._db/616._db, &
                             z1=3._db/40._db, z2=875._db/2244._db, z3=23._db/72._db, &
                             z4=264._db/1955._db, &
                             q1=1._db/6._db, q2=1._db/15._db, q3=1._db/160._db, q4=1._db/144._db
       real(kind=db) :: k1, k2, k3, k4, k5, k6, k7, k8
       real(kind=db) :: R, q, t0, h, tzph, errass, yy_new

       h = hinit
       t0 = 0._db
       yy = x0
       mcycle: do !while ( t0 .lt. periodo )
         attempts: do
           tzph = t0+h
           k1 = rhs(t0,lambda,yy)
           k2 = rhs(t0+h*q1,lambda,yy+k1*q1)
           k3 = rhs(t0+c1*h,lambda,yy+c2*k1+c3*k2)
           k4 = rhs(t0+d1*h,lambda,yy+d2*k1-d3*k2+d4*k3)
           k5 = rhs(t0+d2*h,lambda,yy-e1*k1+e2*k2-e3*k3+e4*k4)
           k6 = rhs(tzph,lambda,yy+f1*k1-8._db*k2+f2*k3-f3*k4+f4*k5)
           k7 = rhs(t0+h*q2,lambda,yy-g1*k1+g2*k2-g3*k3-g4*k4+g5*k5)
           k8 = rhs(tzph,lambda, yy+h1*k1-h2*k2+h3*k3-h4*k4+h5*k5+h6*k7)
           errass = abs(k1*q3+w1*k3-k4*q4+w2*k5+w3*k6-w4*k7-w5*k8)*h
           yy_new = yy+h*(z1*k1+z2*k3+z3*k4+z4*k5+w4*k7+w5*k8)
           R = max(atoll,errass/(atoll+rtoll*max(abs(yy_new),abs(yy))))
           q = max(mins,min(maxs,0.87_db *  R**(-0.2_db) ))
           if ((R .lt. 1._db).or.(h.le.hmin)) then  !! accept step
             t0 = tzph
             h=min(q*h,hmax,periodo-t0)
             yy = yy_new
             exit attempts
           else
             h=min(max(q,1._db)*h,hmin)
             end if
           end do attempts

         if ( t0 .ge. periodo)   exit mcycle
         end do mcycle
       end function d_RKV56_adaptive



   

   end module rk_scalar


   