! StartingPoints-1D Copyright (C) 2013  Marco Spadini
! This file is part of StartingPoints-1D.

! StartingPoints-1D is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! StartingPoints-1D is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with StartingPoints-1D.  If not, see <http://www.gnu.org/licenses/>.
! 
! 
! About the author:
! Marco Spadini                         
! Dipartimento di Matematica e Informatica,
! Via S. Marta 3, 50139 Firenze,       
! ITALY 
! 
! web: http:\\www.dma.unifi.it\~spadini
! e-mail: marco.spadini@unifi.it

program startingPoints
use rk_scalar

  implicit none
  type punto
    real :: x
    real :: l
  end type punto
  interface
    pure function d_deriv(dt,dlambda,dxi)
      integer,parameter :: dbl = selected_real_kind(p=16,r=20)
      real(kind=dbl),intent(in) :: dt,dlambda
      real(kind=dbl),intent(in) :: dxi
      real(kind=dbl) :: d_deriv
    end function d_deriv
    pure function backward(dt,dlambda,dxi)
      integer,parameter :: dbl = selected_real_kind(p=16,r=20)
      real(kind=dbl),intent(in) :: dt,dlambda
      real(kind=dbl),intent(in) :: dxi
      real(kind=dbl) :: backward
    end function backward
  end interface

integer,parameter :: dbl = selected_real_kind(p=16,r=20)
integer,parameter :: MAXITER=50 !! number points per axis
real,allocatable,dimension(:,:) :: valGridF, valGridB !! translation forward and backwards
type(punto),allocatable,dimension(:,:) :: puntiGriglia !! holds grid points

character (len=24) :: arg,fnome
  real :: a, b, c, d, tollr, x, l
  real :: su, giu, dx, sx, bdx, bsx
  real, dimension(2) :: sw , xysize , xysteps 
  integer :: m, i, j, n, nome
  real(kind=dbl) :: Tem
  
!! read command line
call getarg(1,arg)
  read(arg,*) a   !! minimum value for x
call getarg(2,arg)
  read(arg,*) b   !! minimum value for L
call getarg(3,arg)
  read(arg,*) c   !! maximum value for x
call getarg(4,arg)
  read(arg,*) d   !! maximum value for L
call getarg(5,arg)
  read(arg,*) n   !! # of random points (half of it, really)
call getarg(6,arg)
  read(arg,*) tollr  !! tollerance (numerical zero)
call getarg(7,arg)
  read(arg,*) Tem       !! period
call getarg(8,arg)
  read(arg,*) nome       !! output file number
  write(fnome,"(A7,I0.2,A4)") "disegno", nome,".dat"

  allocate(puntiGriglia(0:n,0:n))
  allocate(valGridF(0:n,0:n))
  allocate(valGridB(0:n,0:n))

  sw = (/ a , b /)
  xysize= (/ c-a , d-b /)
  xysteps= xysize/n

  open(unit=1,file=fnome,form='formatted',asynchronous='yes',status='replace')
!$omp parallel private(i,j,su,giu,dx,sx,bdx,bsx,x,l)
!$omp do
  do i=0,n
    l = b + i*xysteps(2)
    do j=0,n
      x = a + j*xysteps(1)
      puntiGriglia(i,j) = punto(x,l)
      valGridF(i,j) = f(x,l)
      valGridB(i,j) = backf(x,l)
    end do
  end do
!$omp end do
!$omp do
  do i=0,n
    do j= 0, n-1
      su = puntiGriglia(i,j+1) % x + tollr
      giu = puntiGriglia(i,j) % x - tollr
      dx = valGridF(i,j)
      sx = valGridF(i,j+1)
      bdx = valGridB(i,j)
      bsx = valGridB(i,j+1)
      if ( ( ((dx.gt.su).and.(sx.lt.su)) .or. ((dx.lt.su).and.(dx.gt.giu)) ) .or. &
           ( ((bdx.gt.su).and.(bsx.lt.su)) .or. ((bdx.lt.su).and.(bdx.gt.giu)) ) ) then
        write(1,*) puntigriglia(i,j) % x,  puntigriglia(i,j) % l
      end if
    end do
  end do
!$omp end do 
!$omp end parallel
  close(1)
  deallocate(puntiGriglia)
  deallocate(valGridF)
  deallocate(valGridB)

contains

pure elemental function f(xi,lambda)
  real,intent(IN) :: xi, lambda
  real :: f
  f = d_rkdp_adaptive(d_deriv,real(xi,kind=dbl),Tem,real(lambda,kind=dbl))
end function f

pure elemental function backf(xi,lambda)
  real,intent(IN) :: xi, lambda
  real :: backf
  backf = d_rkdp_adaptive(backward,real(xi,kind=dbl),Tem,real(lambda,kind=dbl))
end function backf

end program startingPoints


pure function backward(dt,dlambda,dxi)
  interface
    pure function d_deriv(dt,dlambda,dxi)
      integer,parameter :: dbl = selected_real_kind(p=16,r=20)
      real(kind=dbl),intent(in) :: dt,dlambda
      real(kind=dbl),intent(in) :: dxi
      real(kind=dbl) :: d_deriv
    end function d_deriv
  end interface
  integer,parameter :: dbl = selected_real_kind(p=16,r=20)
  real(kind=dbl),intent(in) :: dt, dlambda
  real(kind=dbl),intent(in) :: dxi
  real(kind=dbl) :: backward
      
  backward = -d_deriv(-dt,dlambda,dxi)
  end function backward
