! StartingPoints-1D Copyright (C) 2013  Marco Spadini
! This file is part of StartingPoints-1D.

! StartingPoints-1D is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! StartingPoints-1D is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with StartingPoints-1D.  If not, see <http://www.gnu.org/licenses/>.
! 
! 
! About the author:
! Marco Spadini                         
! Dipartimento di Matematica e Informatica,
! Via S. Marta 3, 50139 Firenze,       
! ITALY 
! 
! web: http:\\www.dma.unifi.it\~spadini
! e-mail: marco.spadini@unifi.it

program startingPoints
use rk_scalar

  implicit none
  interface
    pure function d_deriv(dt,dlambda,dxi)
      integer,parameter :: dbl = selected_real_kind(p=16,r=20)
      real(kind=dbl),intent(in) :: dt,dlambda
      real(kind=dbl),intent(in) :: dxi
      real(kind=dbl) :: d_deriv
    end function d_deriv
    pure function backward(dt,dlambda,dxi)
      integer,parameter :: dbl = selected_real_kind(p=16,r=20)
      real(kind=dbl),intent(in) :: dt,dlambda
      real(kind=dbl),intent(in) :: dxi
      real(kind=dbl) :: backward
    end function backward
  end interface
integer,parameter :: dbl = selected_real_kind(p=16,r=20)
integer,parameter :: MAXITER=50 !! maximum number of bisection iteration
real,allocatable,dimension(:,:) :: puntiAcaso, puntiAcaso2 !! holds random points
real,allocatable,dimension(:) :: valoriAcaso, valoriAcaso2 !! holds images of random points

character (len=24) :: arg,fnome
  real :: a, b, c, d, tollr, vs, v2s
  real :: v , v1 , v2
  real, dimension(2) :: sw , xysize, minixy
  real, dimension(2) :: xr , x0 , x1 , x2
  integer :: m, i, n, nome
  real(kind=dbl) :: Tem
  
!! read command line
call getarg(1,arg)
  read(arg,*) a   !! minimum value for x
call getarg(2,arg)
  read(arg,*) b   !! minimum value for L
call getarg(3,arg)
  read(arg,*) c   !! maximum value for x
call getarg(4,arg)
  read(arg,*) d   !! maximum value for L
call getarg(5,arg)
  read(arg,*) n   !! # of random points (half of it, really)
call getarg(6,arg)
  read(arg,*) tollr  !! tollerance (numerical zero)
call getarg(7,arg)
  read(arg,*) Tem       !! period
call getarg(8,arg)
  read(arg,*) nome       !! output file number
  write(fnome,"(A7,I0.2,A4)") "disegno", nome,".dat"

  allocate(puntiAcaso(2,0:n))
  allocate(puntiAcaso2(2,0:n))
  allocate(valoriAcaso(0:n))
  allocate(valoriAcaso2(0:n))
  call random_number(puntiAcaso)
  call random_number(puntiAcaso2)

  sw = (/ a , b /)
  xysize= (/ c-a , d-b /)
  minixy = xysize*0.1

  open(unit=1,file=fnome,form='formatted',asynchronous='yes',status='replace')
!$omp parallel private(i,m,x0,x1,x2,v,v1,v2,vs,v2s)
!$omp workshare
  puntiAcaso2 = (puntiAcaso2-0.5)
!$omp end workshare
!$omp do schedule(runtime)
  do i=0,n
    puntiAcaso(:,i) = sw+xysize*puntiAcaso(:,i)
    puntiAcaso2(:,i) = puntiAcaso(:,i)+puntiAcaso2(:,i)*minixy 
    valoriAcaso(i) = f(puntiAcaso(1,i),puntiAcaso(2,i))
    valoriAcaso2(i) = f(puntiAcaso2(1,i),puntiAcaso2(2,i))
    if (abs(valoriAcaso(i))<=tollr) then
        write(1,*) puntiAcaso(:,i)
    endif
    if (abs(valoriAcaso2(i))<=tollr) then
        write(1,*) puntiAcaso2(:,i)
    endif
  end do
!$omp end do nowait
!$omp do schedule(runtime)
  do i=0,n
      if (sign(1.,valoriAcaso(i)) .ne. sign(1.,valoriAcaso2(i))) then  !! bisection method starts
        x0=puntiAcaso(:,i)
        v=valoriAcaso(i)
        x1=puntiAcaso2(:,i)
        vs = sign(1.,v)
biszn:     do m=1,MAXITER
             x2=(x0+x1)/2.
             v2=f(x2(1),x2(2))
             v2s = sign(1.,v2)
             if (abs(v2) < tollr) then 
               write(1,*) x2(1), x2(2)
               exit biszn
             elseif (vs .ne. v2s) then
               x1=x2
             else
               x0=x2
               vs=v2s
             endif
           end do biszn        
        endif
      end do
!$omp end do
!$omp do schedule(runtime)
do i=0,n
    valoriAcaso(i) = backf(puntiAcaso(1,i),puntiAcaso(2,i))
    valoriAcaso2(i) = backf(puntiAcaso2(1,i),puntiAcaso2(2,i))
    if (abs(valoriAcaso(i))<=tollr) then
        write(1,*) puntiAcaso(1,i),puntiAcaso(2,i)
    endif
    if (abs(valoriAcaso2(i))<=tollr) then
        write(1,*) puntiAcaso2(1,i),puntiAcaso2(2,i)
    endif
  end do
!$omp end do nowait
!$omp do  schedule(runtime)
  do i=0,n
      if (sign(1.,valoriAcaso(i)) .ne. sign(1.,valoriAcaso2(i))) then  !! bisection method starts
        x0=puntiAcaso(:,i)
        v=valoriAcaso(i)
        x1=puntiAcaso2(:,i)
        vs = sign(1.,v)
biszn2:    do m=1,MAXITER
             x2=(x0+x1)/2.
             v2=backf(x2(1),x2(2))
             v2s = sign(1.,v2)
             if (abs(v2) < tollr) then 
               write(1,*) x2(1), x2(2)
               exit biszn2
             elseif (vs .ne. v2s) then
               x1=x2
             else
               x0=x2
               vs=v2s
             endif
           end do biszn2        
        endif
     end do
!$omp end do
!$omp end parallel
  deallocate(puntiAcaso)
  deallocate(valoriAcaso)
  deallocate(puntiAcaso2)
  deallocate(valoriAcaso2)
  close(1)


contains

pure elemental function f(xi,lambda)
   real,intent(IN) :: xi, lambda
   real :: f

  f = real(xi,kind=dbl) - d_rkdp_adaptive(d_deriv,real(xi,kind=dbl),Tem,real(lambda,kind=dbl))
end function f

pure elemental function backf(xi,lambda)
   real,intent(IN) :: xi, lambda
   real :: backf

  backf = real(xi,kind=dbl) - d_rkdp_adaptive(backward,real(xi,kind=dbl),Tem,real(lambda,kind=dbl))
end function backf

end program startingPoints


pure function backward(dt,dlambda,dxi)
  interface
    pure function d_deriv(dt,dlambda,dxi)
      integer,parameter :: dbl = selected_real_kind(p=16,r=20)
      real(kind=dbl),intent(in) :: dt,dlambda
      real(kind=dbl),intent(in) :: dxi
      real(kind=dbl) :: d_deriv
    end function d_deriv
  end interface
  integer,parameter :: dbl = selected_real_kind(p=16,r=20)
  real(kind=dbl),intent(in) :: dt, dlambda
  real(kind=dbl),intent(in) :: dxi
  real(kind=dbl) :: backward
      
  backward = -d_deriv(-dt,dlambda,dxi)
  end function backward