#!/usr/bin/python
# StartingPoints-1D Copyright (C) 2013 - 2016  Marco Spadini
#
# StartingPoints-1D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StartingPoints-1D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with StartingPoints-1D.  If not, see <http://www.gnu.org/licenses/>.
# 
# 
# About the author:
# Marco Spadini                         
# Dipartimento di Matematica e Informatica,
# Via S. Marta 3, 50139 Firenze,       
# ITALY 
# 
# web: http:\\www.dma.unifi.it\~spadini
# e-mail: marco.spadini@unifi.it
#
#
#  Description:  
#  Given a parameter-dependent differential equation x'=f(t,x,L), where f 
#  is T-periodic in t, the program attempts to find the pairs (L,x) such 
#  that the solution of the differential equation, corresponding to L and 
#  starting at time t=0 from x, is T-periodic.
#
#  The function f ought to be inserted as a string (using FORTRAN 90/95
#  conventions and intrinsics). The search is limited to (L,x) in a 
#  rectangle (L0,L1)X(x0,x1), where L0, L1, x0 and x1 can be specified by
#  the user using appropriate boxes. The same holds for the period T.
#  The user is also prompted for a tollerance 'toll', which defaults to 
#  1/10000, and represents the 'numerical zero', and for the number of
#  points to be used for the computation. Higher numbers correspond to
#  'denser' graphics. All boxes accept simple expressions, like e/2 or 
#  pi. A word of caution concerns integer division: recall that in 
#  FORTRAN, as in many languages, 1./2. = 0.5 whereas 1/2 = 0.

#The program outputs the graphics in a window (one for each test) and
#saves the raw data in files called "disegno*.dat' whith * replaced by 
#a progressive number (one per each test). The same is true for the function 
#used. The user has the option, on exit, to keep the data (e.g. for further 
#processing).



from Tkinter import *
from math import *
import sys, os, commands
#import numpy as np #worked around this
#global idProva
#global version 
#global ncpu
#global opzioni

version= '0.99'
idProva = 0
ncpu = 1
(stt, opF) = commands.getstatusoutput("gcc -dumpversion")
if stt ==0:
  print "found gcc version ", opF
(stt, output) = commands.getstatusoutput("gfortran -v")
if stt:
  print "gfortran not found: exiting"
  sys.exit(1)
if opF >= '4.6':
  opF = '-Ofast'
else:
  opF = '-O3'
print 'Optimization flag set to ', opF  
  

#tries to determine the number of CPUs
try:
  import multiprocessing
  ncpu = multiprocessing.cpu_count()
except (ImportError,NotImplementedError):
  print 'Failed to determine the number of CPUs'
if (ncpu-1):
  print ('Detected %s CPUs. '% ncpu)+'Enabling OpenMP support.'
  omp = ' -fopenmp'
else:
  print 'Only one CPU detected. OpenMP support disabled.'
  omp =''

#removes stale integrator modules (avoids compatibility issues)
try: 
  os.system('rm -f *.o') 
except:
  pass
try: 
    os.system('rm -f *.mod') 
except:
  pass
try: 
  opzioni = ' %s -march=native -mtune=native %s'% (opF, omp)
  #compiles integrator module for the system in use 
  for i in ['Ode6-scalar.f90','pertODE5-e3.f03','pertODE-exp.f03']:
    os.system('gfortran '+i+opzioni+' -c') 
except:
  print 'Some module compilation has failed. Troubles may arise!'

#changes user-interface appearance during computation
def busy(event=None): 
  comp1.pack(side='left')
  comp.pack_forget()
  q1.pack(side='left')
  q.pack_forget()
  outxt.set('....busy...')
  root.update()
  root.update_idletasks()
        
#sets user-interface back to normal after computation
def notbusy(event=None,error=0): 
  comp1.pack_forget()
  comp.pack(side='left')
  q1.pack_forget()
  q.pack(side='left')
  if error:
    outxt.set('...error...')
  else:  
    outxt.set('...ready...')
  root.update()
  root.update_idletasks()
  

def filtra3(xpt,ypt,namein,nameout):
  orVal = open(namein)
  punti = [map(float,riga.split()) for riga in orVal.readlines()]
  orVal.close()
  #punti = np.loadtxt(namein,ndmin=2) # does not work with older versions of numpy
  lx = xpt/(max(punti[:][0])-min(punti[:][0]))
  ly = ypt/(max(punti[:][1])-min(punti[:][1]))
  
  output = list(set([(round(p[0]*lx)/float(lx),round(p[1]*ly)/float(ly)) for p in punti]))
  
  out = open(nameout,'w')
  for item1,item2 in output:
    print>>out, item1,item2
  out.close()
  return len(output),len(punti)

  
#main -- does all the work:
#gathers and convert data, compile FORTRAN main program for the system in use
#prepares and call Gnuplot plots, and saves data
def calc(event=None):
  global idProva, omp, opzioni
  busy()
  idProva += 1
  try:
    funz = f_entry.get()
    x0 = eval(x0_entry.get())
    x1 = eval(x1_entry.get())
    x00 = str(1.05*x0 - 0.05*x1)
    x11 = str(1.05*x1 - 0.05*x0)
    y0 = eval(y0_entry.get())
    y1 = eval(y1_entry.get())
    y00 = str(1.05*y0 - 0.05*y1)
    y11 = str(1.05*y1 - 0.05*y0)
    toll = str(eval(toll_entry.get()))
    T = str(eval(T_entry.get()))
    if compr.get() == '0':
      n = 6*n_entry.get()+ 600*(fact.get()-1) 
      print "Sampling ",n**2," points"
      n = str(n)
    else:  
      n = str(int(n_entry.get()*1000+100000*(fact.get()-1)))
      print "Sampling ",n," points"
  except:
    print "Funny data"
    notbusy(error=1)
    return 1
  f=open('funzione%s.f90'% idProva,'w')
  f.write("""
 pure function d_deriv(t,l,x)
   implicit none
   integer,parameter :: dbl = selected_real_kind(p=16,r=20)
   real(kind=dbl),intent(IN) :: t
   real(kind=dbl),intent(IN) :: l
   real(kind=dbl),intent(IN) :: x
   real(kind=dbl) :: d_deriv

                """ +" d_deriv ="+funz+"\n end function  d_deriv\n")
  f.close()
  comando = 'gfortran '
  if compr.get() == '0':
    comando += 'pertODE-exp.o '
  else:
    comando += 'pertODE5-e3.o '
  comando += 'funzione%s.f90 Ode6-scalar.o '% idProva
  comando += opzioni+' -o parODE'
  failure=os.system(comando)
  if failure:
     print 'compilation error\n\n'
     notbusy(error=1)
     return 1
  dati = x00+" "+y00+" "+x11+" "+y11+" "+n+" "+toll+" "+T+" "+str(idProva)
  failure=os.system('./parODE '+dati)
  if filtro.get():
    coarseness = n_entry.get()*10+200
    print "Coarseness: ",coarseness
    filtra3(coarseness,coarseness,"disegno%s.dat" % str(idProva).zfill(2), "disegno_f%s.dat" % str(idProva).zfill(2))
  if failure:
     print 'execution failure\n'
     notbusy(error=1)
     return 1
  f = open('disegno.gnuplot', 'w')
  f.write("unset label; set autoscale; ")
  f.write("set terminal wxt enhanced persist; ")
  f.write("set title 'x''="+funz+"' ; ")
  f.write("set yrange ["+str(x0)+":"+str(x1)+"] ; ")
  f.write("set xrange ["+str(y0)+":"+str(y1)+"] ; ")
  f.write("set ylabel 'x({/Symbol l},p_{0},0)' ; ")
  f.write("set xlabel '{/Symbol l}' ; ")
  if filtro.get():
    f.write("plot 'disegno_f%s.dat' using 2:1 notitle with dot;" % str(idProva).zfill(2))
  else:
    f.write("plot 'disegno%s.dat' using 2:1 notitle with dot;" % str(idProva).zfill(2))
  f.close()
  failure=os.system('gnuplot disegno.gnuplot')
  if failure:
     print 'plotting failure'
     notbusy(error=1)
     return 1
  if stampa.get():   
    f = open('disegno_eps.gnuplot', 'w')
    f.write("set terminal postscript eps monochrome enhanced;")
    f.write("set output 'disegno%s.eps' ;"%  str(idProva).zfill(2))
    f.write("unset label; set autoscale; ")
    f.write("unset title; ")
    f.write("set yrange ["+str(x0)+":"+str(x1)+"] ; ")
    f.write("set xrange ["+str(y0)+":"+str(y1)+"] ; ")
    
    f.write("set xlabel font ',30' ; ")
    f.write("set ylabel font ',30' ; ")
    f.write("set xtics font ',18' ; ")
    f.write("set ytics font ',18' ; ")
    
    f.write("set ylabel 'x({/Symbol l},p_{0},0)' ; ")
    f.write("set xlabel '{/Symbol l}' ; ")
    if filtro.get():
      f.write("plot 'disegno_f%s.dat' using 2:1 notitle pt 7 ps 0.4;" % str(idProva).zfill(2))
    else:
      f.write("plot 'disegno%s.dat' using 2:1 notitle pt 7 ps 0.4;" % str(idProva).zfill(2))
    f.write("unset output;")
    f.close()
    failure=os.system('gnuplot disegno_eps.gnuplot')
    if failure:
       print 'plotting failure'
       notbusy(error=1)
       return 1
  notbusy(error=0)

 
#Quits the program
def smetti(event=None):
  if salva.get():
    try: 
      os.system('rm -f disegno*.dat') 
    except:
      print 'disegno*.dat non presente'
  if salvaeps.get():
    try: 
      os.system('rm -f disegno*.eps') 
    except:
      pass
    try: 
      os.system('rm -f funzione*.f90') 
    except:
      print 'funzione*.f90 non presente'
  
  try: 
    os.remove('disegno.gnuplot') 
  except:
    pass
  try: 
    os.remove('disegno_eps.gnuplot') 
  except:
    pass
  try: 
    os.remove('parODE') 
  except:
    pass
  try: 
    os.system('rm -f *.o') 
  except:
    pass
  try:
    os.system('rm -f *.mod') 
  except:
    pass
  print " 'Bye"
  root.quit() 


#interface construction
root = Tk()
root.title('StartingPoints-1D %s'% version)
salva=IntVar()
salvaeps=IntVar()
stampa=IntVar()
filtro=IntVar()
fact=IntVar()
compr=StringVar()
outxt=StringVar()
toptop = Frame(root); toptop.pack(padx=10,pady=10)
Label(toptop, 
text="""Looks for starting points of T-periodic solutions in (L0,L1)X(x0,x1)
of parametrized scalar equation x\'=f(t,x,L), assuming f is T-periodic in \
t""",fg='brown' ).pack(side='top')

top = Frame(root); top.pack()
Label(top, text='f(t,x,L):= ').pack(side='left')
# use get and configure to interact with the texts
f_entry = Entry(top, width=64)
f_entry.pack(side='left')
f_entry.insert('end', 'x+L*sin(t)') #sets default value  

mid = Frame(root); mid.pack()
        
Label(mid, text='  x0 =').pack(side='left')
x0_entry = Entry(mid, width=4)
x0_entry.pack(side='left')
x0_entry.insert('end', '-3.')  #sets default value  

Label(mid, text='  x1 =').pack(side='left')
x1_entry = Entry(mid, width=4)
x1_entry.pack(side='left')
x1_entry.insert('end', '3.')  #sets default value  

Label(mid, text='  L0 =').pack(side='left')
y0_entry = Entry(mid, width=4)
y0_entry.pack(side='left')
y0_entry.insert('end', '0.')  #sets default value  

Label(mid, text='  L1 =').pack(side='left')
y1_entry = Entry(mid, width=4)
y1_entry.pack(side='left')
y1_entry.insert('end', '1.')  #sets default value  


Label(mid, text='  T =').pack(side='left')
T_entry = Entry(mid, width=4)
T_entry.pack(side='left')
T_entry.insert('end', '2.*pi')

Label(mid, text='  toll =').pack(side='left')
toll_entry = Entry(mid, width=4)
toll_entry.pack(side='left')
toll_entry.insert('end', '1.e-4')  #sets default value  


bot = Frame(root); bot.pack()
Label(bot, text="Resolution % = ").pack(side='left')
n_entry = Scale(bot, orient='horizontal', from_=1, to=100, 
                tickinterval=99, resolution=1, length= 300)
n_entry.set(25)  #sets default value                
n_entry.pack(side='left',padx=5)
Label(bot, text="  Factor =").pack(side='left',padx=3)
Spinbox(bot, from_=1, to=9, increment=1, width=1, 
        textvariable=fact).pack(side='left')

bbot = Frame(root)
bbot.pack()
bl = LabelFrame(bbot,text = 'status', width=109)
bl.pack(side='left',padx=20,ipadx=30,ipady=15,fill=Y)
outxt.set('...ready...')
Label(bl,textvariable=outxt,fg="brown").pack(expand='true')
blr = LabelFrame(bbot,text = 'compute',fg='blue',height=35, width=7)
blr.pack(side='left',padx=5,ipadx=10)
subblr = Frame(blr)
subblr.pack(side='left',padx=5,pady=5)
ssubblr = Frame(subblr)
ssubblr.pack(side='top',anchor='nw')
Label(ssubblr,text='Method: ').pack(side='left')
rb1=Radiobutton(ssubblr, text="soft", variable=compr, value='1')
rb1.pack(side="left")
rb2=Radiobutton(ssubblr, text="hard", variable=compr, value='0')
rb2.pack(side="left")
rb1.select()
eps = Checkbutton(subblr,text="Generate eps?",onvalue=1,offvalue=0,variable=stampa)
fltr = Checkbutton(subblr,text="Grid reduce?",onvalue=1,offvalue=0,variable=filtro)
eps.pack(side='top',anchor='sw')
fltr.pack(side='top',anchor='sw')
comp = Button(blr, text=' GO ', relief='raised', cursor="hand1",fg="blue", command=calc)
comp1 = Button(blr, text=' GO ', relief='sunken', cursor="watch",fg="gray",state=DISABLED)
comp.pack(side='left')



bbotf = LabelFrame(bbot,text = 'exit',fg='red',height=35, width=7)
bbotf.pack(padx=5,ipadx=5,ipady=8,fill=Y)
bbotr = Frame(bbotf)
bbotr.pack(side='left')
keepdata = Checkbutton(bbotr,text="Erase data\n on exit?",onvalue=1,offvalue=0,variable=salva)
keepdata.pack(side='top',padx=12)
keepeps = Checkbutton(bbotr,text="Clear eps'?",onvalue=1,offvalue=0,variable=salvaeps)
keepeps.pack(side='top',padx=12)
q = Button(bbotf, text=' QUIT ', relief='raised', cursor="pirate",fg="red", command=smetti)
q1 = Button(bbotf, text=' QUIT ', relief='raised', cursor="watch",fg="gray",state=DISABLED)
q.pack(side='left')
root.protocol('WM_DELETE_WINDOW', smetti)
    

#displays licence agreement
root.lic = Toplevel()
root.lic.title('StartingPoints-1D %s'% version)
testo ="""StartingPoints-1D %s  Copyright (C) 2013-2016  Marco Spadini\n
    
StartingPoints-1D Copyright (C) 2013-2016  Marco Spadini

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

StartingPoints-1D is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with StartingPoints-1D.  If not, see <http://www.gnu.org/licenses/>.


About the author:
Marco Spadini                         
Dipartimento di Matematica e Informatica,
Via S. Marta 3, 50139 Firenze,       
ITALY 

web: http:\\\www.dma.unifi.it\~spadini
e-mail: marco.spadini@unifi.it""" % version
Message(root.lic,aspect=200, justify="left", text=testo).pack(padx=9,pady=10)
def azione(): # destroys window
  root.lic.destroy()
Button(root.lic, text='Close',fg="red",cursor="pirate", command=azione).pack()
root.lic.protocol('WM_DELETE_WINDOW', azione) #intercepts window close event


root.mainloop()

